// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use fayalite::{prelude::*, sim::Simulation};

#[hdl_module(outline_generated)]
pub fn connect_const() {
    #[hdl]
    let o: UInt<8> = m.output();
    connect(o, 5u8);
}

#[test]
fn test_connect_const() {
    let _n = SourceLocation::normalize_files_for_tests();
    let mut sim = Simulation::new(connect_const());
    sim.settle_step();
    let sim_debug = format!("{sim:#?}");
    println!("#######\n{sim_debug}\n#######");
    if sim_debug
        != r#"Simulation {
    state: State {
        insns: Insns {
            state_layout: StateLayout {
                ty: TypeLayout {
                    small_slots: StatePartAllocationLayout<SmallSlots> {
                        len: 0,
                        debug_data: [],
                        ..
                    },
                    big_slots: StatePartAllocationLayout<BigSlots> {
                        len: 2,
                        debug_data: [
                            SlotDebugData {
                                name: "InstantiatedModule(connect_const: connect_const).connect_const::o",
                                ty: UInt<8>,
                            },
                            SlotDebugData {
                                name: "",
                                ty: UInt<8>,
                            },
                        ],
                        ..
                    },
                },
            },
            insns: [
                // at: module-XXXXXXXXXX.rs:1:1
                Const {
                    dest: StatePartIndex<BigSlots>(1), // SlotDebugData { name: "", ty: UInt<8> },
                    value: 5,
                },
                // at: module-XXXXXXXXXX.rs:3:1
                Copy {
                    dest: StatePartIndex<BigSlots>(0), // SlotDebugData { name: "InstantiatedModule(connect_const: connect_const).connect_const::o", ty: UInt<8> },
                    src: StatePartIndex<BigSlots>(1), // SlotDebugData { name: "", ty: UInt<8> },
                },
                // at: module-XXXXXXXXXX.rs:1:1
                Return,
            ],
            ..
        },
        pc: 2,
        small_slots: StatePart {
            value: [],
        },
        big_slots: StatePart {
            value: [
                5,
                5,
            ],
        },
    },
    base_module: CompiledModule {
        module_io: [
            CompiledValue {
                layout: CompiledTypeLayout {
                    ty: UInt<8>,
                    layout: TypeLayout {
                        small_slots: StatePartAllocationLayout<SmallSlots> {
                            len: 0,
                            debug_data: [],
                            ..
                        },
                        big_slots: StatePartAllocationLayout<BigSlots> {
                            len: 1,
                            debug_data: [
                                SlotDebugData {
                                    name: "InstantiatedModule(connect_const: connect_const).connect_const::o",
                                    ty: UInt<8>,
                                },
                            ],
                            ..
                        },
                    },
                    body: Scalar,
                },
                range: TypeIndexRange {
                    small_slots: StatePartIndexRange<SmallSlots> { start: 0, len: 0 },
                    big_slots: StatePartIndexRange<BigSlots> { start: 0, len: 1 },
                },
                write: None,
            },
        ],
    },
    base_module_io_ty: connect_const {
        o: UInt<8>,
    },
}"# {
        panic!();
    }
}

#[hdl_module(outline_generated)]
pub fn mod1_child() {
    #[hdl]
    let i: UInt<4> = m.input();
    #[hdl]
    let o: SInt<2> = m.output();
    #[hdl]
    let i2: SInt<2> = m.input();
    #[hdl]
    let o2: UInt<4> = m.output();
    connect(o, i.cast_to_static());
    connect(o2, i2.cast_to_static());
    #[hdl]
    if i.cmp_gt(5_hdl_u4) {
        connect(o2, 0xF_hdl_u4);
    }
}

#[hdl_module(outline_generated)]
pub fn mod1() {
    #[hdl]
    let child = instance(mod1_child());
    #[hdl]
    let o: mod1_child = m.output(Expr::ty(child));
    connect(o, child);
}

#[test]
fn test_mod1() {
    let _n = SourceLocation::normalize_files_for_tests();
    let mut sim = Simulation::new(mod1());
    sim.settle_step();
    let sim_debug = format!("{sim:#?}");
    println!("#######\n{sim_debug}\n#######");
    if sim_debug
        != r#"Simulation {
    state: State {
        insns: Insns {
            state_layout: StateLayout {
                ty: TypeLayout {
                    small_slots: StatePartAllocationLayout<SmallSlots> {
                        len: 0,
                        debug_data: [],
                        ..
                    },
                    big_slots: StatePartAllocationLayout<BigSlots> {
                        len: 17,
                        debug_data: [
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::o.i",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::o.o",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::o.i2",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::o.o2",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::child.i",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::child.o",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::child.i2",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1: mod1).mod1::child.o2",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i2",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o2",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "",
                                ty: SInt<2>,
                            },
                            SlotDebugData {
                                name: "",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "",
                                ty: UInt<4>,
                            },
                            SlotDebugData {
                                name: "",
                                ty: Bool,
                            },
                            SlotDebugData {
                                name: "",
                                ty: UInt<4>,
                            },
                        ],
                        ..
                    },
                },
            },
            insns: [
                // at: module-XXXXXXXXXX.rs:4:1
                Copy {
                    dest: StatePartIndex<BigSlots>(6), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.i2", ty: SInt<2> },
                    src: StatePartIndex<BigSlots>(2), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::o.i2", ty: SInt<2> },
                },
                Copy {
                    dest: StatePartIndex<BigSlots>(4), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.i", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(0), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::o.i", ty: UInt<4> },
                },
                // at: module-XXXXXXXXXX.rs:2:1
                Copy {
                    dest: StatePartIndex<BigSlots>(10), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i2", ty: SInt<2> },
                    src: StatePartIndex<BigSlots>(6), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.i2", ty: SInt<2> },
                },
                Copy {
                    dest: StatePartIndex<BigSlots>(8), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(4), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.i", ty: UInt<4> },
                },
                // at: module-XXXXXXXXXX-2.rs:1:1
                Const {
                    dest: StatePartIndex<BigSlots>(16), // SlotDebugData { name: "", ty: UInt<4> },
                    value: 15,
                },
                Const {
                    dest: StatePartIndex<BigSlots>(14), // SlotDebugData { name: "", ty: UInt<4> },
                    value: 5,
                },
                CmpLt {
                    dest: StatePartIndex<BigSlots>(15), // SlotDebugData { name: "", ty: Bool },
                    lhs: StatePartIndex<BigSlots>(14), // SlotDebugData { name: "", ty: UInt<4> },
                    rhs: StatePartIndex<BigSlots>(8), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i", ty: UInt<4> },
                },
                CastToUInt {
                    dest: StatePartIndex<BigSlots>(13), // SlotDebugData { name: "", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(10), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i2", ty: SInt<2> },
                    dest_width: 4,
                },
                // at: module-XXXXXXXXXX-2.rs:7:1
                Copy {
                    dest: StatePartIndex<BigSlots>(11), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o2", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(13), // SlotDebugData { name: "", ty: UInt<4> },
                },
                // at: module-XXXXXXXXXX-2.rs:8:1
                BranchIfZero {
                    target: 11,
                    value: StatePartIndex<BigSlots>(15), // SlotDebugData { name: "", ty: Bool },
                },
                // at: module-XXXXXXXXXX-2.rs:9:1
                Copy {
                    dest: StatePartIndex<BigSlots>(11), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o2", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(16), // SlotDebugData { name: "", ty: UInt<4> },
                },
                // at: module-XXXXXXXXXX.rs:2:1
                Copy {
                    dest: StatePartIndex<BigSlots>(7), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.o2", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(11), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o2", ty: UInt<4> },
                },
                // at: module-XXXXXXXXXX.rs:4:1
                Copy {
                    dest: StatePartIndex<BigSlots>(3), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::o.o2", ty: UInt<4> },
                    src: StatePartIndex<BigSlots>(7), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.o2", ty: UInt<4> },
                },
                // at: module-XXXXXXXXXX-2.rs:1:1
                CastToSInt {
                    dest: StatePartIndex<BigSlots>(12), // SlotDebugData { name: "", ty: SInt<2> },
                    src: StatePartIndex<BigSlots>(8), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::i", ty: UInt<4> },
                    dest_width: 2,
                },
                // at: module-XXXXXXXXXX-2.rs:6:1
                Copy {
                    dest: StatePartIndex<BigSlots>(9), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o", ty: SInt<2> },
                    src: StatePartIndex<BigSlots>(12), // SlotDebugData { name: "", ty: SInt<2> },
                },
                // at: module-XXXXXXXXXX.rs:2:1
                Copy {
                    dest: StatePartIndex<BigSlots>(5), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.o", ty: SInt<2> },
                    src: StatePartIndex<BigSlots>(9), // SlotDebugData { name: "InstantiatedModule(mod1.child: mod1_child).mod1_child::o", ty: SInt<2> },
                },
                // at: module-XXXXXXXXXX.rs:4:1
                Copy {
                    dest: StatePartIndex<BigSlots>(1), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::o.o", ty: SInt<2> },
                    src: StatePartIndex<BigSlots>(5), // SlotDebugData { name: "InstantiatedModule(mod1: mod1).mod1::child.o", ty: SInt<2> },
                },
                // at: module-XXXXXXXXXX.rs:1:1
                Return,
            ],
            ..
        },
        pc: 17,
        small_slots: StatePart {
            value: [],
        },
        big_slots: StatePart {
            value: [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                5,
                0,
                15,
            ],
        },
    },
    base_module: CompiledModule {
        module_io: [
            CompiledValue {
                layout: CompiledTypeLayout {
                    ty: Bundle {
                        #[hdl(flip)] /* offset = 0 */
                        i: UInt<4>,
                        /* offset = 4 */
                        o: SInt<2>,
                        #[hdl(flip)] /* offset = 6 */
                        i2: SInt<2>,
                        /* offset = 8 */
                        o2: UInt<4>,
                    },
                    layout: TypeLayout {
                        small_slots: StatePartAllocationLayout<SmallSlots> {
                            len: 0,
                            debug_data: [],
                            ..
                        },
                        big_slots: StatePartAllocationLayout<BigSlots> {
                            len: 4,
                            debug_data: [
                                SlotDebugData {
                                    name: "InstantiatedModule(mod1: mod1).mod1::o.i",
                                    ty: UInt<4>,
                                },
                                SlotDebugData {
                                    name: "InstantiatedModule(mod1: mod1).mod1::o.o",
                                    ty: SInt<2>,
                                },
                                SlotDebugData {
                                    name: "InstantiatedModule(mod1: mod1).mod1::o.i2",
                                    ty: SInt<2>,
                                },
                                SlotDebugData {
                                    name: "InstantiatedModule(mod1: mod1).mod1::o.o2",
                                    ty: UInt<4>,
                                },
                            ],
                            ..
                        },
                    },
                    body: Bundle {
                        fields: [
                            CompiledBundleField {
                                offset: TypeIndex {
                                    small_slots: StatePartIndex<SmallSlots>(0),
                                    big_slots: StatePartIndex<BigSlots>(0),
                                },
                                ty: CompiledTypeLayout {
                                    ty: UInt<4>,
                                    layout: TypeLayout {
                                        small_slots: StatePartAllocationLayout<SmallSlots> {
                                            len: 0,
                                            debug_data: [],
                                            ..
                                        },
                                        big_slots: StatePartAllocationLayout<BigSlots> {
                                            len: 1,
                                            debug_data: [
                                                SlotDebugData {
                                                    name: "",
                                                    ty: UInt<4>,
                                                },
                                            ],
                                            ..
                                        },
                                    },
                                    body: Scalar,
                                },
                            },
                            CompiledBundleField {
                                offset: TypeIndex {
                                    small_slots: StatePartIndex<SmallSlots>(0),
                                    big_slots: StatePartIndex<BigSlots>(1),
                                },
                                ty: CompiledTypeLayout {
                                    ty: SInt<2>,
                                    layout: TypeLayout {
                                        small_slots: StatePartAllocationLayout<SmallSlots> {
                                            len: 0,
                                            debug_data: [],
                                            ..
                                        },
                                        big_slots: StatePartAllocationLayout<BigSlots> {
                                            len: 1,
                                            debug_data: [
                                                SlotDebugData {
                                                    name: "",
                                                    ty: SInt<2>,
                                                },
                                            ],
                                            ..
                                        },
                                    },
                                    body: Scalar,
                                },
                            },
                            CompiledBundleField {
                                offset: TypeIndex {
                                    small_slots: StatePartIndex<SmallSlots>(0),
                                    big_slots: StatePartIndex<BigSlots>(2),
                                },
                                ty: CompiledTypeLayout {
                                    ty: SInt<2>,
                                    layout: TypeLayout {
                                        small_slots: StatePartAllocationLayout<SmallSlots> {
                                            len: 0,
                                            debug_data: [],
                                            ..
                                        },
                                        big_slots: StatePartAllocationLayout<BigSlots> {
                                            len: 1,
                                            debug_data: [
                                                SlotDebugData {
                                                    name: "",
                                                    ty: SInt<2>,
                                                },
                                            ],
                                            ..
                                        },
                                    },
                                    body: Scalar,
                                },
                            },
                            CompiledBundleField {
                                offset: TypeIndex {
                                    small_slots: StatePartIndex<SmallSlots>(0),
                                    big_slots: StatePartIndex<BigSlots>(3),
                                },
                                ty: CompiledTypeLayout {
                                    ty: UInt<4>,
                                    layout: TypeLayout {
                                        small_slots: StatePartAllocationLayout<SmallSlots> {
                                            len: 0,
                                            debug_data: [],
                                            ..
                                        },
                                        big_slots: StatePartAllocationLayout<BigSlots> {
                                            len: 1,
                                            debug_data: [
                                                SlotDebugData {
                                                    name: "",
                                                    ty: UInt<4>,
                                                },
                                            ],
                                            ..
                                        },
                                    },
                                    body: Scalar,
                                },
                            },
                        ],
                    },
                },
                range: TypeIndexRange {
                    small_slots: StatePartIndexRange<SmallSlots> { start: 0, len: 0 },
                    big_slots: StatePartIndexRange<BigSlots> { start: 0, len: 4 },
                },
                write: None,
            },
        ],
    },
    base_module_io_ty: mod1 {
        o: mod1_child {
            i: UInt<4>,
            o: SInt<2>,
            i2: SInt<2>,
            o2: UInt<4>,
        },
    },
}"# {
        panic!();
    }
}
